import 'dart:async';
import 'dart:io';
import 'dart:typed_data';
import 'package:camera/camera.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:get/get.dart';
import 'package:image/image.dart' as img;
import 'package:pedantic/pedantic.dart';

import '../../../configs/application.dart';
import '../../app/controllers/app_controller.dart';
import '../../display_photo/views/display_photo_view.dart';
import '../views/id_card_scanner_view.dart';
import '../../../routes/app_pages.dart';
import '../../../utils/logger.dart';

class IdCardScannerController extends GetxController {
  late List<CameraDescription> cameras;

  late IdCardScannerViewArguments arguments;
  final scaffoldKey = GlobalKey<ScaffoldState>();
  final imageKey = GlobalKey();
  late CameraController cameraCtrller;
  bool cameraInitial = false;
  RxBool loading = false.obs;
  RxInt duration = 5.obs;
  Timer? _timer;
  @override
  void onInit() {
    arguments = Get.arguments ??
        IdCardScannerViewArguments(
          title: 'front'.tr.toLowerCase(),
          isBackSide: false,
        );
    init();
    super.onInit();
  }

  void init() async {
    cameras = await availableCameras();
    cameraCtrller = CameraController(
      cameras[0],
      ResolutionPreset.high,
      enableAudio: false,
    );
    unawaited(cameraCtrller.initialize().then((_) {
      cameraInitial = true;
      if (Application.config.autoORC) {
        startTimer();
      }
      update();
    }));
  }

  void startTimer() {
    _timer = Timer.periodic(const Duration(seconds: 1), (timer) {
      duration.value = duration.value - 1;
      if (duration.value == 0) {
        _timer?.cancel();
        takePhoto(arguments.isBackSide);
      }
    });
  }

  Future<void> setFlashMode(FlashMode mode) async {
    try {
      await cameraCtrller.setFlashMode(mode);
    } on CameraException catch (e) {
      UtilLogger.log(e);
    }
  }

  @override
  void onClose() {
    cameraCtrller.dispose();
    _timer?.cancel();
  }

  Future<Uint8List> _loadImageSource(XFile xFileImage) async {
    var src = img.decodeImage(File(xFileImage.path).readAsBytesSync());
    final rect = imageKey.globalPaintBounds;
    final width = rect.width;
    final height = rect.width;
    var widthRatio = src!.width / Get.width;
    var heightRatio = src.height / Get.height;
    final newRect = Rect.fromLTWH(
      (src.width - width * widthRatio) / 2,
      (src.height - height * heightRatio) / 2,
      width * widthRatio,
      height * heightRatio,
    );
    UtilLogger.log('width ${rect.width}, height ${rect.height}');
    UtilLogger.log('crop width ${newRect.width},crop height ${newRect.height}');
    src = img.copyCrop(
      src,
      (newRect.left).floor(),
      (newRect.top).floor(),
      newRect.width.floor(),
      newRect.height.floor(),
    );
    return Uint8List.fromList(img.encodeJpg(src));
  }

  void takePhoto(bool isBackSide) async {
    loading.value = true;
    var image = await takePicture();
    if (image != null) {
      try {
        var parseImage = await _loadImageSource(image);
        loading.value = false;
        Get.find<AppController>()
            .updateCard(parseImage, isBackside: isBackSide);
        Get.offAndToNamed(
          Routes.DISPLAY_PHOTO,
          arguments: DisplayPhotoViewArguments(
            isBackSide: isBackSide,
            image: parseImage,
            btnText: (isBackSide ? 'confirm' : 'taske_backside').tr,
          ),
        );
      } catch (e) {
        loading.value = false;
        showInSnackBar(e.toString());
      }
    } else {
      showInSnackBar('Not capture in camera');
      loading.value = false;
    }
  }

  void showInSnackBar(String message) {
    ScaffoldMessenger.of(scaffoldKey.currentState!.context).showSnackBar(
      SnackBar(
        content: Text(message),
      ),
    );
  }

  Future<XFile?> takePicture() async {
    final CameraController? cameraController = cameraCtrller;
    if (cameraController == null || !cameraController.value.isInitialized) {
      showInSnackBar('Error: select a camera first.');
      return null;
    }

    if (cameraController.value.isTakingPicture) {
      return null;
    }

    try {
      var file = await cameraController.takePicture();
      return file;
    } on CameraException catch (e) {
      _showCameraException(e);
      return null;
    }
  }

  void _showCameraException(CameraException e) {
    showInSnackBar('Error: ${e.code}\n${e.description}');
  }
}

extension GlobalKeyExtension on GlobalKey {
  Rect get globalPaintBounds {
    final renderObject = currentContext?.findRenderObject() as RenderBox;
    var translation = renderObject.getTransformTo(null).getTranslation();
    return renderObject.paintBounds.shift(Offset(translation.x, translation.y));
  }
}
